package com.confidosoft.health_kit

import android.app.Activity
import android.content.Context
import android.content.Intent
import android.util.Log
import androidx.annotation.NonNull
import com.google.android.gms.auth.api.signin.GoogleSignIn
import com.google.android.gms.auth.api.signin.GoogleSignInOptions
import com.google.android.gms.fitness.Fitness
import com.google.android.gms.fitness.FitnessOptions
import com.google.android.gms.fitness.data.*
import com.google.android.gms.fitness.request.DataReadRequest
import com.google.android.gms.fitness.request.SessionReadRequest
import com.google.android.gms.fitness.result.DataReadResponse
import com.google.android.gms.fitness.result.SessionReadResponse
import io.flutter.embedding.engine.plugins.FlutterPlugin
import io.flutter.embedding.engine.plugins.activity.ActivityAware
import io.flutter.embedding.engine.plugins.activity.ActivityPluginBinding
import io.flutter.plugin.common.BinaryMessenger
import io.flutter.plugin.common.MethodCall
import io.flutter.plugin.common.MethodChannel
import io.flutter.plugin.common.MethodChannel.MethodCallHandler
import io.flutter.plugin.common.MethodChannel.Result
import io.flutter.plugin.common.PluginRegistry
import io.flutter.plugin.common.PluginRegistry.Registrar
import java.util.concurrent.TimeUnit

// Created by Rashmi Tank on 2021-04-27.

/** HealthKitPlugin */
class HealthKitPlugin : FlutterPlugin, MethodCallHandler, ActivityAware, PluginRegistry.ActivityResultListener {
    /// The MethodChannel that will the communication between Flutter and native Android
    ///
    /// This local reference serves to register the plugin with the Flutter Engine and unregister it
    /// when the Flutter Engine is detached from the Activity
    private var channel: MethodChannel? = null
    private val channelName = "health_kit"
    private var flutterPluginBinding: FlutterPlugin.FlutterPluginBinding? = null
    private var mActivity: Activity? = null

    interface OAuthPermissionsListener {
        fun onOAuthPermissionsResult(resultCode: Int)
    }

    private val oAuthPermissionListeners = mutableListOf<OAuthPermissionsListener>()


    companion object {
        private const val TAG = "HealthKit"
        private const val TAG_UNSUPPORTED = "unsupported"
        private const val GOOGLE_FIT_PERMISSIONS_REQUEST_CODE = 1234

        @JvmStatic
        @SuppressWarnings("deprecation")
        fun registerWith(registrar: Registrar) {
            val plugin = HealthKitPlugin()
            registrar.addActivityResultListener(plugin)
            registrar.activity()?.let { plugin.setupMethodChannelOnActivityAttached(it, plugin, registrar.messenger(), registrar.context()) }
        }
    }

    /**
     * This method registers method call handler to the channel
     */
    private fun setupMethodChannelOnActivityAttached(activity: Activity, plugin: HealthKitPlugin, messenger: BinaryMessenger, context: Context) {
        mActivity = activity

        plugin.channel = MethodChannel(messenger, channelName)
        plugin.channel?.setMethodCallHandler(plugin)
    }

    override fun onAttachedToEngine(@NonNull binding: FlutterPlugin.FlutterPluginBinding) {
        this.flutterPluginBinding = binding
    }

    override fun onDetachedFromEngine(@NonNull binding: FlutterPlugin.FlutterPluginBinding) {
        this.flutterPluginBinding = null
    }

    override fun onDetachedFromActivity() {
        channel?.setMethodCallHandler(null)
        channel = null
        mActivity = null
    }

    override fun onReattachedToActivityForConfigChanges(binding: ActivityPluginBinding) {
        onAttachedToActivity(binding)
    }

    override fun onAttachedToActivity(binding: ActivityPluginBinding) {
        binding.addActivityResultListener(this)

        flutterPluginBinding?.binaryMessenger?.let { setupMethodChannelOnActivityAttached(binding.activity, this, it, binding.activity) }
    }

    override fun onDetachedFromActivityForConfigChanges() {
        onDetachedFromActivity()
    }


    override fun onMethodCall(@NonNull call: MethodCall, @NonNull result: Result) {
        try {
            when (call.method) {
                "hasPermissions" -> {
                    hasPermissions(getFitnessOptions(call), result)
                }
                "requestPermissions" -> {
                    requestPermissions(getFitnessOptions(call), result)
                }
                "read" -> {
                    val request = ReadRequest.fromCall(call)
                    read(request, result)
                }
                "revokePermissions" -> revokePermissions(result)
                else -> result.notImplemented()
            }
        } catch (e: Throwable) {
            result.error(TAG, e.message, null)
        }
    }

    override fun onActivityResult(requestCode: Int, resultCode: Int, data: Intent?): Boolean {
        if (requestCode == GOOGLE_FIT_PERMISSIONS_REQUEST_CODE) {
            Log.d(TAG, "Google Fit Access Granted!")
            oAuthPermissionListeners.toList()
                    .forEach { it.onOAuthPermissionsResult(resultCode) }
            return true
        }
        return false
    }

    /**
     * Creates instance of FitnessOptions from data types
     *
     * @param call object of method call
     * @return returns FitnessOptions
     */
    private fun getFitnessOptions(call: MethodCall): FitnessOptions {
        val listTypes: List<Type> = call.argument<List<String>>("types")
                ?.mapNotNull { type -> type.fromDartType() }
                ?: throw Exception("types is not defined")

        val builder = FitnessOptions.builder()
        for (type in listTypes) {
            builder.addDataType(type.dataType)
        }
        return builder.build()
    }

    /**
     * Checks whether permission is given for data types and returns true/false
     *
     * @param fitnessOptions
     * @param result
     */
    private fun hasPermissions(fitnessOptions: FitnessOptions, result: Result) {
        if (hasOAuthPermission(fitnessOptions)) {
            result.success(true)
        } else {
            result.success(false)
        }
    }

    /**
     * It requests permission for given data types and returns result true/false
     *
     * @param fitnessOptions
     * @param result
     */
    private fun requestPermissions(fitnessOptions: FitnessOptions, result: Result) {
        requestOAuthPermissions(fitnessOptions, {
            result.success(true)
        }, {
            result.success(false)
        })
    }

    /**
     * It requests permission for given data types and returns result
     *
     * @param fitnessOptions
     * @param onSuccess
     * @param onError
     */
    private fun requestOAuthPermissions(fitnessOptions: FitnessOptions, onSuccess: () -> Unit, onError: () -> Unit) {
        if (hasOAuthPermission(fitnessOptions)) {
            onSuccess()
            return
        }

        oAuthPermissionListeners.add(object : OAuthPermissionsListener {
            override fun onOAuthPermissionsResult(resultCode: Int) {
                oAuthPermissionListeners.remove(this)

                if (resultCode == Activity.RESULT_OK) {
                    onSuccess()
                } else {
                    onError()
                }
            }
        })

        mActivity?.let {
            GoogleSignIn.requestPermissions(
                    it,
                    GOOGLE_FIT_PERMISSIONS_REQUEST_CODE,
                    GoogleSignIn.getLastSignedInAccount(it),
                    fitnessOptions)
        }
    }

    /**
     * Checks whether oAuthPermission is available
     *
     * @param fitnessOptions
     * @return
     */
    private fun hasOAuthPermission(fitnessOptions: FitnessOptions): Boolean {
        return GoogleSignIn.hasPermissions(GoogleSignIn.getLastSignedInAccount(mActivity), fitnessOptions)
    }

    /**
     * Reads Sample or Activity data based on data types
     *
     * @param request
     * @param result
     */
    private fun read(request: ReadRequest<*>, result: Result) {
        val options = FitnessOptions.builder()
                .addDataType(request.type.dataType)
                .build()

        requestOAuthPermissions(options, {
            when (request) {
                is ReadRequest.Sample -> readSample(request, result)
                is ReadRequest.Activity -> readSession(request, result)
            }
        }, {
            result.error(TAG, "User denied permission access", null)
        })
    }

    /**
     * Reads sample data
     *
     * @param request
     * @param result
     */
    private fun readSample(request: ReadRequest<Type.Sample>, result: Result) {
        Log.d(TAG, "readSample: ${request.type.dataType.name}")
        var readRequest: DataReadRequest? = null

        if (request.type.dataType.name == "com.google.step_count.delta") {
            // add condition of aggregate estimated data type for step count
            val estimatedStepsDelta: DataSource = DataSource.Builder()
                    .setDataType(request.type.dataType)
                    .setType(DataSource.TYPE_DERIVED)
                    .setStreamName("estimated_steps")
                    .setAppPackageName("com.google.android.gms")
                    .build()

            readRequest = DataReadRequest.Builder()
                    .aggregate(estimatedStepsDelta)
                    .also { builder ->
                        when (request.limit != null) {
                            true -> builder.setLimit(request.limit)
                            else -> builder.bucketByTime(1, TimeUnit.DAYS)
                        }
                    }
                    .setTimeRange(request.dateFrom.time, request.dateTo.time, TimeUnit.MILLISECONDS)
                    .enableServerQueries()
                    .build()
        } else {
            readRequest = DataReadRequest.Builder()
                    .read(request.type.dataType)
                    .also { builder ->
                        when (request.limit != null) {
                            true -> builder.setLimit(request.limit)
                            else -> builder.bucketByTime(1, TimeUnit.DAYS)
                        }
                    }
                    .setTimeRange(request.dateFrom.time, request.dateTo.time, TimeUnit.MILLISECONDS)
                    .enableServerQueries()
                    .build()
        }

        mActivity?.let { mActivity ->
            GoogleSignIn.getLastSignedInAccount(mActivity)?.let { googleSignInAccount ->
                readRequest?.let { dataReadRequest ->
                    Fitness.getHistoryClient(mActivity, googleSignInAccount)
                            .readData(dataReadRequest)
                            .addOnSuccessListener { response ->
                                onSuccess(response, result)
                            }
                            .addOnFailureListener { e ->
                                Log.i("GoogleFit", "addOnFailureListener: $e")
                                result.error(TAG, e.message, null)
                            }
                            .addOnCanceledListener {
                                Log.i("GoogleFit", "addOnCanceledListener")
                                result.error(TAG, "GoogleFit Cancelled", null)
                            }
                }

            }
        }
    }

    /**
     * Extracts sample data on read success from google fit
     *
     * @param response
     * @param result
     */
    private fun onSuccess(response: DataReadResponse, result: Result) {
        (response.dataSets + response.buckets.flatMap { it.dataSets })
                .filterNot { it.isEmpty }
                .flatMap { it.dataPoints }
                .map(::dataPointToMap)
                .let(result::success)
    }

    @Suppress("IMPLICIT_CAST_TO_ANY")
    private fun dataPointToMap(dataPoint: DataPoint): Map<String, Any> {
        val field = dataPoint.dataType.fields.first()
        val source = dataPoint.originalDataSource.streamName

        return mapOf(
                "value" to dataPoint.getValue(field).let { value ->
                    when (value.format) {
                        Field.FORMAT_FLOAT -> value.asFloat()
                        Field.FORMAT_INT32 -> value.asInt()
                        else -> {
                            //NA
                        }
                    }
                },
                "date_from" to dataPoint.getStartTime(TimeUnit.MILLISECONDS),
                "date_to" to dataPoint.getEndTime(TimeUnit.MILLISECONDS),
                "source" to source,
                "user_entered" to (source == "user_input")
        )
    }

    /**
     * Reads Session data
     *
     * @param request
     * @param result
     */
    private fun readSession(request: ReadRequest<Type.Activity>, result: Result) {
        Log.d(TAG, "readSession: ${request.type.activity}")

        val readRequest = SessionReadRequest.Builder()
                .read(request.type.dataType)
                .setTimeInterval(request.dateFrom.time, request.dateTo.time, TimeUnit.MILLISECONDS)
                .readSessionsFromAllApps()
                .enableServerQueries()
                .build()

        mActivity?.let { mActivity ->
            GoogleSignIn.getLastSignedInAccount(mActivity)?.let { googleSignInAccount ->
                Fitness.getSessionsClient(mActivity, googleSignInAccount)
                        .readSession(readRequest)
                        .addOnSuccessListener { response -> onSuccess(request, response, result) }
                        .addOnFailureListener { e -> result.error(TAG, e.message, null) }
                        .addOnCanceledListener { result.error(TAG, "GoogleFit Cancelled", null) }
            }
        }
    }

    /**
     * Extract Session data on read success from google fit
     *
     * @param request
     * @param response
     * @param result
     */
    private fun onSuccess(request: ReadRequest<Type.Activity>, response: SessionReadResponse, result: Result) {
        response.sessions.filter { request.type.activity == it.activity }
                .let { list ->
                    when (request.limit != null) {
                        true -> list.takeLast(request.limit)
                        else -> list
                    }
                }
                .map { session -> sessionToMap(session, response.getDataSet(session)) }
                .let(result::success)
    }

    /**
     * Converts session data to Map and returns it
     *
     * @param session
     * @param dataSets
     * @return
     */
    private fun sessionToMap(session: Session, dataSets: List<DataSet>): Map<String, Any> {
        // from all data points find the top used streamName
        val source = dataSets.asSequence()
                .filterNot { it.isEmpty }
                .flatMap { it.dataPoints.asSequence() }
                .filterNot { it.originalDataSource.streamName.isNullOrEmpty() }
                .groupingBy { it.originalDataSource.streamName }
                .eachCount()
                .maxByOrNull { it.value }
                ?.key ?: session.name ?: ""

        return mapOf(
                "value" to session.getValue(),
                "date_from" to session.getStartTime(TimeUnit.MILLISECONDS),
                "date_to" to session.getEndTime(TimeUnit.MILLISECONDS),
                "source" to source,
                "user_entered" to (source == "user_input")
        )
    }

    /**
     * Revokes granted permission to the app
     *
     * @param result
     */
    private fun revokePermissions(result: Result) {
        val fitnessOptions = FitnessOptions.builder()
                .build()

        if (!hasOAuthPermission(fitnessOptions)) {
            result.success(null)
            return
        }

        mActivity?.let { mActivity ->
            GoogleSignIn.getLastSignedInAccount(mActivity)?.let { googleSignInAccount ->
                Fitness.getConfigClient(mActivity, googleSignInAccount)
                        .disableFit()
                        .continueWithTask {
                            val signInOptions = GoogleSignInOptions.Builder()
                                    .addExtension(fitnessOptions)
                                    .build()
                            GoogleSignIn.getClient(mActivity, signInOptions)
                                    .revokeAccess()
                        }
                        .addOnSuccessListener { result.success(null) }
                        .addOnFailureListener { e ->
                            if (!hasOAuthPermission(fitnessOptions)) {
                                result.success(null)
                            } else {
                                result.error(TAG, e.message, null)
                            }
                        }
            }
        }
    }

}
